@@ -24,9 +24,14 @@
 // NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
+use std::mem::MaybeUninit;
+
 use ring::aead;
 use ring::hkdf;
 
+use libc::c_int;
+use libc::c_void;
+
 use crate::Error;
 use crate::Result;
 
@@ -68,11 +73,13 @@ pub enum Algorithm {
 }
 
 impl Algorithm {
-    fn get_ring_aead(self) -> &'static aead::Algorithm {
+    fn get_evp_aead(self) -> *const EVP_AEAD {
         match self {
-            Algorithm::AES128_GCM => &aead::AES_128_GCM,
-            Algorithm::AES256_GCM => &aead::AES_256_GCM,
-            Algorithm::ChaCha20_Poly1305 => &aead::CHACHA20_POLY1305,
+            Algorithm::AES128_GCM => unsafe { EVP_aead_aes_128_gcm() },
+            Algorithm::AES256_GCM => unsafe { EVP_aead_aes_256_gcm() },
+            Algorithm::ChaCha20_Poly1305 => unsafe {
+                EVP_aead_chacha20_poly1305()
+            },
         }
     }
 
@@ -93,7 +100,11 @@ impl Algorithm {
     }
 
     pub fn key_len(self) -> usize {
-        self.get_ring_aead().key_len()
+        match self {
+            Algorithm::AES128_GCM => 16,
+            Algorithm::AES256_GCM => 32,
+            Algorithm::ChaCha20_Poly1305 => 32,
+        }
     }
 
     pub fn tag_len(self) -> usize {
@@ -101,20 +112,28 @@ impl Algorithm {
             return 0;
         }
 
-        self.get_ring_aead().tag_len()
+        match self {
+            Algorithm::AES128_GCM => 16,
+            Algorithm::AES256_GCM => 16,
+            Algorithm::ChaCha20_Poly1305 => 16,
+        }
     }
 
     pub fn nonce_len(self) -> usize {
-        self.get_ring_aead().nonce_len()
+        match self {
+            Algorithm::AES128_GCM => 12,
+            Algorithm::AES256_GCM => 12,
+            Algorithm::ChaCha20_Poly1305 => 12,
+        }
     }
 }
 
 pub struct Open {
     alg: Algorithm,
 
-    hp_key: aead::quic::HeaderProtectionKey,
+    ctx: EVP_AEAD_CTX,
 
-    key: aead::LessSafeKey,
+    hp_key: aead::quic::HeaderProtectionKey,
 
     nonce: Vec<u8>,
 }
@@ -124,20 +143,17 @@ impl Open {
         alg: Algorithm, key: &[u8], iv: &[u8], hp_key: &[u8],
     ) -> Result<Open> {
         Ok(Open {
+            alg,
+
+            ctx: make_aead_ctx(alg, key)?,
+
             hp_key: aead::quic::HeaderProtectionKey::new(
                 alg.get_ring_hp(),
                 hp_key,
             )
             .map_err(|_| Error::CryptoFail)?,
 
-            key: aead::LessSafeKey::new(
-                aead::UnboundKey::new(alg.get_ring_aead(), key)
-                    .map_err(|_| Error::CryptoFail)?,
-            ),
-
             nonce: Vec::from(iv),
-
-            alg,
         })
     }
 
@@ -149,9 +165,9 @@ impl Open {
         let mut iv = vec![0; nonce_len];
         let mut pn_key = vec![0; key_len];
 
-        derive_pkt_key(aead, &secret, &mut key)?;
-        derive_pkt_iv(aead, &secret, &mut iv)?;
-        derive_hdr_key(aead, &secret, &mut pn_key)?;
+        derive_pkt_key(aead, secret, &mut key)?;
+        derive_pkt_iv(aead, secret, &mut iv)?;
+        derive_hdr_key(aead, secret, &mut pn_key)?;
 
         Open::new(aead, &key, &iv, &pn_key)
     }
@@ -163,16 +179,37 @@ impl Open {
             return Ok(buf.len());
         }
 
+        let tag_len = self.alg().tag_len();
+
+        let mut out_len = match buf.len().checked_sub(tag_len) {
+            Some(n) => n,
+            None => return Err(Error::CryptoFail),
+        };
+
+        let max_out_len = out_len;
+
         let nonce = make_nonce(&self.nonce, counter);
 
-        let ad = aead::Aad::from(ad);
+        let rc = unsafe {
+            EVP_AEAD_CTX_open(
+                &self.ctx,          // ctx
+                buf.as_mut_ptr(),   // out
+                &mut out_len,       // out_len
+                max_out_len,        // max_out_len
+                nonce[..].as_ptr(), // nonce
+                nonce.len(),        // nonce_len
+                buf.as_ptr(),       // inp
+                buf.len(),          // in_len
+                ad.as_ptr(),        // ad
+                ad.len(),           // ad_len
+            )
+        };
 
-        let plain = self
-            .key
-            .open_in_place(nonce, ad, buf)
-            .map_err(|_| Error::CryptoFail)?;
+        if rc != 1 {
+            return Err(Error::CryptoFail);
+        }
 
-        Ok(plain.len())
+        Ok(out_len)
     }
 
     pub fn new_mask(&self, sample: &[u8]) -> Result<[u8; 5]> {
@@ -196,9 +233,9 @@ impl Open {
 pub struct Seal {
     alg: Algorithm,
 
-    hp_key: aead::quic::HeaderProtectionKey,
+    ctx: EVP_AEAD_CTX,
 
-    key: aead::LessSafeKey,
+    hp_key: aead::quic::HeaderProtectionKey,
 
     nonce: Vec<u8>,
 }
@@ -208,20 +245,17 @@ impl Seal {
         alg: Algorithm, key: &[u8], iv: &[u8], hp_key: &[u8],
     ) -> Result<Seal> {
         Ok(Seal {
+            alg,
+
+            ctx: make_aead_ctx(alg, key)?,
+
             hp_key: aead::quic::HeaderProtectionKey::new(
                 alg.get_ring_hp(),
                 hp_key,
             )
             .map_err(|_| Error::CryptoFail)?,
 
-            key: aead::LessSafeKey::new(
-                aead::UnboundKey::new(alg.get_ring_aead(), key)
-                    .map_err(|_| Error::CryptoFail)?,
-            ),
-
             nonce: Vec::from(iv),
-
-            alg,
         })
     }
 
@@ -233,40 +267,66 @@ impl Seal {
         let mut iv = vec![0; nonce_len];
         let mut pn_key = vec![0; key_len];
 
-        derive_pkt_key(aead, &secret, &mut key)?;
-        derive_pkt_iv(aead, &secret, &mut iv)?;
-        derive_hdr_key(aead, &secret, &mut pn_key)?;
+        derive_pkt_key(aead, secret, &mut key)?;
+        derive_pkt_iv(aead, secret, &mut iv)?;
+        derive_hdr_key(aead, secret, &mut pn_key)?;
 
         Seal::new(aead, &key, &iv, &pn_key)
     }
 
     pub fn seal_with_u64_counter(
-        &self, counter: u64, ad: &[u8], buf: &mut [u8],
-    ) -> Result<()> {
+        &self, counter: u64, ad: &[u8], buf: &mut [u8], in_len: usize,
+        extra_in: Option<&[u8]>,
+    ) -> Result<usize> {
         if cfg!(feature = "fuzzing") {
-            return Ok(());
+            if let Some(extra) = extra_in {
+                buf[in_len..in_len + extra.len()].copy_from_slice(extra);
+                return Ok(in_len + extra.len());
+            }
+
+            return Ok(in_len);
+        }
+
+        let tag_len = self.alg().tag_len();
+
+        let mut out_tag_len = tag_len;
+
+        let (extra_in_ptr, extra_in_len) = match extra_in {
+            Some(v) => (v.as_ptr(), v.len()),
+
+            None => (std::ptr::null(), 0),
+        };
+
+        // Make sure all the outputs combined fit in the buffer.
+        if in_len + tag_len + extra_in_len > buf.len() {
+            return Err(Error::CryptoFail);
         }
 
         let nonce = make_nonce(&self.nonce, counter);
 
-        let ad = aead::Aad::from(ad);
+        let rc = unsafe {
+            EVP_AEAD_CTX_seal_scatter(
+                &self.ctx,                  // ctx
+                buf.as_mut_ptr(),           // out
+                buf[in_len..].as_mut_ptr(), // out_tag
+                &mut out_tag_len,           // out_tag_len
+                tag_len + extra_in_len,     // max_out_tag_len
+                nonce[..].as_ptr(),         // nonce
+                nonce.len(),                // nonce_len
+                buf.as_ptr(),               // inp
+                in_len,                     // in_len
+                extra_in_ptr,               // extra_in
+                extra_in_len,               // extra_in_len
+                ad.as_ptr(),                // ad
+                ad.len(),                   // ad_len
+            )
+        };
 
-        let tag_len = self.alg().tag_len();
+        if rc != 1 {
+            return Err(Error::CryptoFail);
+        }
 
-        let in_out_len =
-            buf.len().checked_sub(tag_len).ok_or(Error::CryptoFail)?;
-
-        let (in_out, tag_out) = buf.split_at_mut(in_out_len);
-
-        let tag = self
-            .key
-            .seal_in_place_separate_tag(nonce, ad, in_out)
-            .map_err(|_| Error::CryptoFail)?;
-
-        // Append the AEAD tag to the end of the sealed buffer.
-        tag_out.copy_from_slice(tag.as_ref());
-
-        Ok(())
+        Ok(in_len + out_tag_len)
     }
 
     pub fn new_mask(&self, sample: &[u8]) -> Result<[u8; 5]> {
@@ -297,7 +357,7 @@ pub fn derive_initial_key_material(
     let key_len = aead.key_len();
     let nonce_len = aead.nonce_len();
 
-    let initial_secret = derive_initial_secret(&cid, version);
+    let initial_secret = derive_initial_secret(cid, version);
 
     // Client.
     let mut client_key = vec![0; key_len];
@@ -418,6 +478,31 @@ pub fn derive_pkt_iv(
     hkdf_expand_label(&secret, LABEL, &mut out[..nonce_len])
 }
 
+fn make_aead_ctx(alg: Algorithm, key: &[u8]) -> Result<EVP_AEAD_CTX> {
+    let mut ctx = MaybeUninit::uninit();
+
+    let ctx = unsafe {
+        let aead = alg.get_evp_aead();
+
+        let rc = EVP_AEAD_CTX_init(
+            ctx.as_mut_ptr(),
+            aead,
+            key.as_ptr(),
+            alg.key_len(),
+            alg.tag_len(),
+            std::ptr::null_mut(),
+        );
+
+        if rc != 1 {
+            return Err(Error::CryptoFail);
+        }
+
+        ctx.assume_init()
+    };
+
+    Ok(ctx)
+}
+
 fn hkdf_expand_label(
     prk: &hkdf::Prk, label: &[u8], out: &mut [u8],
 ) -> Result<()> {
@@ -436,9 +521,9 @@ fn hkdf_expand_label(
     Ok(())
 }
 
-fn make_nonce(iv: &[u8], counter: u64) -> aead::Nonce {
+fn make_nonce(iv: &[u8], counter: u64) -> [u8; aead::NONCE_LEN] {
     let mut nonce = [0; aead::NONCE_LEN];
-    nonce.copy_from_slice(&iv);
+    nonce.copy_from_slice(iv);
 
     // XOR the last bytes of the IV with the counter. This is equivalent to
     // left-padding the counter with zero bytes.
@@ -446,7 +531,7 @@ fn make_nonce(iv: &[u8], counter: u64) -> aead::Nonce {
         *a ^= b;
     }
 
-    aead::Nonce::assume_unique_for_key(nonce)
+    nonce
 }
 
 // The ring HKDF expand() API does not accept an arbitrary output length, so we
@@ -460,6 +545,49 @@ impl hkdf::KeyType for ArbitraryOutputLen {
     }
 }
 
+#[allow(non_camel_case_types)]
+#[repr(transparent)]
+struct EVP_AEAD(c_void);
+
+// NOTE: This structure is copied from <openssl/aead.h> in order to be able to
+// statically allocate it. While it is not often modified upstream, it needs to
+// be kept in sync.
+#[repr(C)]
+struct EVP_AEAD_CTX {
+    aead: libc::uintptr_t,
+    opaque: [u8; 580],
+    alignment: u64,
+    tag_len: u8,
+}
+
+extern {
+    // EVP_AEAD
+    fn EVP_aead_aes_128_gcm() -> *const EVP_AEAD;
+
+    fn EVP_aead_aes_256_gcm() -> *const EVP_AEAD;
+
+    fn EVP_aead_chacha20_poly1305() -> *const EVP_AEAD;
+
+    // EVP_AEAD_CTX
+    fn EVP_AEAD_CTX_init(
+        ctx: *mut EVP_AEAD_CTX, aead: *const EVP_AEAD, key: *const u8,
+        key_len: usize, tag_len: usize, engine: *mut c_void,
+    ) -> c_int;
+
+    fn EVP_AEAD_CTX_open(
+        ctx: *const EVP_AEAD_CTX, out: *mut u8, out_len: *mut usize,
+        max_out_len: usize, nonce: *const u8, nonce_len: usize, inp: *const u8,
+        in_len: usize, ad: *const u8, ad_len: usize,
+    ) -> c_int;
+
+    fn EVP_AEAD_CTX_seal_scatter(
+        ctx: *const EVP_AEAD_CTX, out: *mut u8, out_tag: *mut u8,
+        out_tag_len: *mut usize, max_out_tag_len: usize, nonce: *const u8,
+        nonce_len: usize, inp: *const u8, in_len: usize, extra_in: *const u8,
+        extra_in_len: usize, ad: *const u8, ad_len: usize,
+    ) -> c_int;
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
