@@ -42,10 +42,15 @@ use crate::minmax;
 use crate::packet;
 use crate::ranges;
 
-// Loss Recovery
-const PACKET_THRESHOLD: u64 = 3;
+#[cfg(feature = "qlog")]
+use qlog::events::EventData;
 
-const TIME_THRESHOLD: f64 = 9.0 / 8.0;
+// Loss Recovery
+const INITIAL_PACKET_THRESHOLD: u64 = 3;
+
+const MAX_PACKET_THRESHOLD: u64 = 20;
+
+const INITIAL_TIME_THRESHOLD: f64 = 9.0 / 8.0;
 
 const GRANULARITY: Duration = Duration::from_millis(1);
 
@@ -64,6 +69,10 @@ const MINIMUM_WINDOW_PACKETS: usize = 2;
 
 const LOSS_REDUCTION_FACTOR: f64 = 0.5;
 
+const PACING_MULTIPLIER: f64 = 1.25;
+
+const MICROS_PER_SEC: u64 = 1_000_000;
+
 pub struct Recovery {
     loss_detection_timer: Option<Instant>,
 
@@ -97,6 +106,8 @@ pub struct Recovery {
 
     pub lost_count: usize,
 
+    pub lost_spurious_count: usize,
+
     pub loss_probes: [usize; packet::EPOCH_COUNT],
 
     in_flight_count: [usize; packet::EPOCH_COUNT],
@@ -105,6 +116,10 @@ pub struct Recovery {
 
     delivery_rate: delivery_rate::Rate,
 
+    pkt_thresh: u64,
+
+    time_thresh: f64,
+
     // Congestion control.
     cc_ops: &'static CongestionControlOps,
 
@@ -120,6 +135,8 @@ pub struct Recovery {
 
     bytes_sent: usize,
 
+    pub bytes_lost: u64,
+
     congestion_recovery_start_time: Option<Instant>,
 
     max_datagram_size: usize,
@@ -133,6 +150,12 @@ pub struct Recovery {
     pacing_rate: u64,
 
     last_packet_scheduled_time: Option<Instant>,
+
+    // RFC6937 PRR.
+    prr: prr::PRR,
+
+    #[cfg(feature = "qlog")]
+    qlog_metrics: QlogMetrics,
 }
 
 impl Recovery {
@@ -148,7 +171,7 @@ impl Recovery {
 
             largest_sent_pkt: [0; packet::EPOCH_COUNT],
 
-            latest_rtt: Duration::new(0, 0),
+            latest_rtt: Duration::ZERO,
 
             // This field should be initialized to `INITIAL_RTT` for the initial
             // PTO calculation, but it also needs to be an `Option` to track
@@ -156,13 +179,13 @@ impl Recovery {
             // handled by the `rtt()` method instead.
             smoothed_rtt: None,
 
-            minmax_filter: minmax::Minmax::new(Duration::new(0, 0)),
+            minmax_filter: minmax::Minmax::new(Duration::ZERO),
 
-            min_rtt: Duration::new(0, 0),
+            min_rtt: Duration::ZERO,
 
             rttvar: INITIAL_RTT / 2,
 
-            max_ack_delay: Duration::new(0, 0),
+            max_ack_delay: Duration::ZERO,
 
             loss_time: [None; packet::EPOCH_COUNT],
 
@@ -173,6 +196,7 @@ impl Recovery {
             acked: [Vec::new(), Vec::new(), Vec::new()],
 
             lost_count: 0,
+            lost_spurious_count: 0,
 
             loss_probes: [0; packet::EPOCH_COUNT],
 
@@ -181,6 +205,10 @@ impl Recovery {
             congestion_window: config.max_send_udp_payload_size *
                 INITIAL_WINDOW_PACKETS,
 
+            pkt_thresh: INITIAL_PACKET_THRESHOLD,
+
+            time_thresh: INITIAL_TIME_THRESHOLD,
+
             bytes_in_flight: 0,
 
             ssthresh: std::usize::MAX,
@@ -191,6 +219,8 @@ impl Recovery {
 
             bytes_sent: 0,
 
+            bytes_lost: 0,
+
             congestion_recovery_start_time: None,
 
             max_datagram_size: config.max_send_udp_payload_size,
@@ -208,9 +238,18 @@ impl Recovery {
             pacing_rate: 0,
 
             last_packet_scheduled_time: None,
+
+            prr: prr::PRR::default(),
+
+            #[cfg(feature = "qlog")]
+            qlog_metrics: QlogMetrics::default(),
         }
     }
 
+    pub fn on_init(&mut self) {
+        (self.cc_ops.on_init)(self);
+    }
+
     pub fn on_packet_sent(
         &mut self, mut pkt: Sent, epoch: packet::Epoch,
         handshake_status: HandshakeStatus, now: Instant, trace_id: &str,
@@ -239,6 +278,8 @@ impl Recovery {
 
             self.on_packet_sent_cc(sent_bytes, now);
 
+            self.prr.on_packet_sent(sent_bytes);
+
             self.set_loss_detection_timer(handshake_status, now);
         }
 
@@ -253,8 +294,10 @@ impl Recovery {
         // Pacing: Set the pacing rate if CC doesn't do its own.
         if !(self.cc_ops.has_custom_pacing)() {
             if let Some(srtt) = self.smoothed_rtt {
-                let rate = (self.congestion_window as u64 * 1000000) /
-                    srtt.as_micros() as u64;
+                let cwnd = self.congestion_window as u64;
+                let rate =
+                    (cwnd * MICROS_PER_SEC) as f64 / srtt.as_micros() as f64;
+                let rate = (rate * PACING_MULTIPLIER) as u64;
                 self.set_pacing_rate(rate);
             }
         }
@@ -298,7 +341,7 @@ impl Recovery {
         self.last_packet_scheduled_time = match self.last_packet_scheduled_time {
             Some(last_scheduled_time) => {
                 let interval: u64 =
-                    (packet_size as u64 * 1000000) / self.pacing_rate;
+                    (packet_size as u64 * MICROS_PER_SEC) / self.pacing_rate;
                 let interval = Duration::from_micros(interval);
                 let next_schedule_time = last_scheduled_time + interval;
                 Some(cmp::max(now, next_schedule_time))
@@ -340,24 +383,52 @@ impl Recovery {
 
         let mut newly_acked = Vec::new();
 
+        let mut undo_cwnd = false;
+
+        let max_rtt = cmp::max(self.latest_rtt, self.rtt());
+
         // Detect and mark acked packets, without removing them from the sent
         // packets list.
         for r in ranges.iter() {
-            let lowest_acked = r.start;
-            let largest_acked = r.end - 1;
+            let lowest_acked_in_block = r.start;
+            let largest_acked_in_block = r.end - 1;
 
             let unacked_iter = self.sent[epoch]
                 .iter_mut()
                 // Skip packets that precede the lowest acked packet in the block.
-                .skip_while(|p| p.pkt_num < lowest_acked)
+                .skip_while(|p| p.pkt_num < lowest_acked_in_block)
                 // Skip packets that follow the largest acked packet in the block.
-                .take_while(|p| p.pkt_num <= largest_acked)
+                .take_while(|p| p.pkt_num <= largest_acked_in_block)
                 // Skip packets that have already been acked or lost.
-                .filter(|p| p.time_acked.is_none() && p.time_lost.is_none());
+                .filter(|p| p.time_acked.is_none());
 
             for unacked in unacked_iter {
                 unacked.time_acked = Some(now);
 
+                // Check if acked packet was already declared lost.
+                if unacked.time_lost.is_some() {
+                    // Calculate new packet reordering threshold.
+                    let pkt_thresh =
+                        self.largest_acked_pkt[epoch] - unacked.pkt_num + 1;
+                    let pkt_thresh = cmp::min(MAX_PACKET_THRESHOLD, pkt_thresh);
+
+                    self.pkt_thresh = cmp::max(self.pkt_thresh, pkt_thresh);
+
+                    // Calculate new time reordering threshold.
+                    let loss_delay = max_rtt.mul_f64(self.time_thresh);
+                    if now.duration_since(unacked.time_sent) > loss_delay {
+                        // TODO: do time threshold update
+                        self.time_thresh = 5_f64 / 4_f64;
+                    }
+
+                    if unacked.in_flight {
+                        undo_cwnd = true;
+                    }
+
+                    self.lost_spurious_count += 1;
+                    continue;
+                }
+
                 if unacked.ack_eliciting {
                     has_ack_eliciting = true;
                 }
@@ -371,7 +442,7 @@ impl Recovery {
                     self.in_flight_count[epoch] =
                         self.in_flight_count[epoch].saturating_sub(1);
 
-                    self.delivery_rate.on_packet_acked(&unacked, now);
+                    self.delivery_rate.on_packet_acked(unacked, now);
                 }
 
                 newly_acked.push(Acked {
@@ -386,6 +457,11 @@ impl Recovery {
             }
         }
 
+        // Undo congestion window update.
+        if undo_cwnd {
+            (self.cc_ops.rollback)(self);
+        }
+
         self.delivery_rate.estimate();
 
         if newly_acked.is_empty() {
@@ -414,7 +490,7 @@ impl Recovery {
 
         self.set_loss_detection_timer(handshake_status, now);
 
-        self.drain_packets(epoch);
+        self.drain_packets(epoch, now);
 
         Ok(())
     }
@@ -521,7 +597,9 @@ impl Recovery {
             return std::usize::MAX;
         }
 
-        self.congestion_window.saturating_sub(self.bytes_in_flight)
+        // Open more space (snd_cnt) for PRR when allowed.
+        self.congestion_window.saturating_sub(self.bytes_in_flight) +
+            self.prr.snd_cnt
     }
 
     pub fn rtt(&self) -> Duration {
@@ -597,7 +675,6 @@ impl Recovery {
         let mut time = self.loss_time[epoch];
 
         // Iterate over all packet number spaces starting from Handshake.
-        #[allow(clippy::needless_range_loop)]
         for e in packet::EPOCH_HANDSHAKE..packet::EPOCH_COUNT {
             let new_time = self.loss_time[e];
 
@@ -684,7 +761,7 @@ impl Recovery {
         self.loss_time[epoch] = None;
 
         let loss_delay =
-            cmp::max(self.latest_rtt, self.rtt()).mul_f64(TIME_THRESHOLD);
+            cmp::max(self.latest_rtt, self.rtt()).mul_f64(self.time_thresh);
 
         // Minimum time of kGranularity before packets are deemed lost.
         let loss_delay = cmp::max(loss_delay, GRANULARITY);
@@ -706,7 +783,7 @@ impl Recovery {
         for unacked in unacked_iter {
             // Mark packet as lost, or set time when it should be marked.
             if unacked.time_sent <= lost_send_time ||
-                largest_acked >= unacked.pkt_num + PACKET_THRESHOLD
+                largest_acked >= unacked.pkt_num + self.pkt_thresh
             {
                 self.lost[epoch].append(&mut unacked.frames);
 
@@ -743,14 +820,16 @@ impl Recovery {
             }
         }
 
+        self.bytes_lost += lost_bytes as u64;
+
         if let Some(pkt) = largest_lost_pkt {
             self.on_packets_lost(lost_bytes, &pkt, epoch, now);
         }
 
-        self.drain_packets(epoch);
+        self.drain_packets(epoch, now);
     }
 
-    fn drain_packets(&mut self, epoch: packet::Epoch) {
+    fn drain_packets(&mut self, epoch: packet::Epoch, now: Instant) {
         let mut lowest_non_expired_pkt_index = self.sent[epoch].len();
 
         // In order to avoid removing elements from the middle of the list
@@ -764,6 +843,13 @@ impl Recovery {
 
         // First, find the first element that is neither acked nor lost.
         for (i, pkt) in self.sent[epoch].iter().enumerate() {
+            if let Some(time_lost) = pkt.time_lost {
+                if time_lost + self.rtt() > now {
+                    lowest_non_expired_pkt_index = i;
+                    break;
+                }
+            }
+
             if pkt.time_acked.is_none() && pkt.time_lost.is_none() {
                 lowest_non_expired_pkt_index = i;
                 break;
@@ -840,22 +926,18 @@ impl Recovery {
     }
 
     #[cfg(feature = "qlog")]
-    pub fn to_qlog(&self) -> qlog::event::Event {
-        // QVis can't use all these fields and they can be large.
-        qlog::event::Event::metrics_updated(
-            Some(self.min_rtt.as_millis() as u64),
-            Some(self.rtt().as_millis() as u64),
-            Some(self.latest_rtt.as_millis() as u64),
-            Some(self.rttvar.as_millis() as u64),
-            None, // delay
-            None, // probe_count
-            Some(self.cwnd() as u64),
-            Some(self.bytes_in_flight as u64),
-            None, // ssthresh
-            None, // packets_in_flight
-            None, // in_recovery
-            None, // pacing_rate
-        )
+    pub fn maybe_qlog(&mut self) -> Option<EventData> {
+        let qlog_metrics = QlogMetrics {
+            min_rtt: self.min_rtt,
+            smoothed_rtt: self.rtt(),
+            latest_rtt: self.latest_rtt,
+            rttvar: self.rttvar,
+            cwnd: self.cwnd() as u64,
+            bytes_in_flight: self.bytes_in_flight as u64,
+            ssthresh: self.ssthresh as u64,
+        };
+
+        self.qlog_metrics.maybe_update(qlog_metrics)
     }
 }
 
@@ -889,6 +971,8 @@ impl FromStr for CongestionControlAlgorithm {
 }
 
 pub struct CongestionControlOps {
+    pub on_init: fn(r: &mut Recovery),
+
     pub on_packet_sent: fn(r: &mut Recovery, sent_bytes: usize, now: Instant),
 
     pub on_packet_acked:
@@ -905,9 +989,12 @@ pub struct CongestionControlOps {
 
     pub checkpoint: fn(r: &mut Recovery),
 
-    pub rollback: fn(r: &mut Recovery),
+    pub rollback: fn(r: &mut Recovery) -> bool,
 
     pub has_custom_pacing: fn() -> bool,
+
+    pub debug_fmt:
+        fn(r: &Recovery, formatter: &mut std::fmt::Formatter) -> std::fmt::Result,
 }
 
 impl From<CongestionControlAlgorithm> for &'static CongestionControlOps {
@@ -965,6 +1052,9 @@ impl std::fmt::Debug for Recovery {
             write!(f, "hystart={:?} ", self.hystart)?;
         }
 
+        // CC-specific debug info
+        (self.cc_ops.debug_fmt)(self, f)?;
+
         Ok(())
     }
 }
@@ -1056,6 +1146,111 @@ fn sub_abs(lhs: Duration, rhs: Duration) -> Duration {
     }
 }
 
+// We don't need to log all qlog metrics every time there is a recovery event.
+// Instead, we can log only the MetricsUpdated event data fields that we care
+// about, only when they change. To support this, the QLogMetrics structure
+// keeps a running picture of the fields.
+#[derive(Default)]
+#[cfg(feature = "qlog")]
+struct QlogMetrics {
+    min_rtt: Duration,
+    smoothed_rtt: Duration,
+    latest_rtt: Duration,
+    rttvar: Duration,
+    cwnd: u64,
+    bytes_in_flight: u64,
+    ssthresh: u64,
+}
+
+#[cfg(feature = "qlog")]
+impl QlogMetrics {
+    // Make a qlog event if the latest instance of QlogMetrics is different.
+    //
+    // This function diffs each of the fields. A qlog MetricsUpdated event is
+    // only generated if at least one field is different. Where fields are
+    // different, the qlog event contains the latest value.
+    fn maybe_update(&mut self, latest: Self) -> Option<EventData> {
+        let mut emit_event = false;
+
+        let new_min_rtt = if self.min_rtt != latest.min_rtt {
+            self.min_rtt = latest.min_rtt;
+            emit_event = true;
+            Some(latest.min_rtt.as_secs_f32() * 1000.0)
+        } else {
+            None
+        };
+
+        let new_smoothed_rtt = if self.smoothed_rtt != latest.smoothed_rtt {
+            self.smoothed_rtt = latest.smoothed_rtt;
+            emit_event = true;
+            Some(latest.smoothed_rtt.as_secs_f32() * 1000.0)
+        } else {
+            None
+        };
+
+        let new_latest_rtt = if self.latest_rtt != latest.latest_rtt {
+            self.latest_rtt = latest.latest_rtt;
+            emit_event = true;
+            Some(latest.latest_rtt.as_secs_f32() * 1000.0)
+        } else {
+            None
+        };
+
+        let new_rttvar = if self.rttvar != latest.rttvar {
+            self.rttvar = latest.rttvar;
+            emit_event = true;
+            Some(latest.rttvar.as_secs_f32() * 1000.0)
+        } else {
+            None
+        };
+
+        let new_cwnd = if self.cwnd != latest.cwnd {
+            self.cwnd = latest.cwnd;
+            emit_event = true;
+            Some(latest.cwnd)
+        } else {
+            None
+        };
+
+        let new_bytes_in_flight =
+            if self.bytes_in_flight != latest.bytes_in_flight {
+                self.bytes_in_flight = latest.bytes_in_flight;
+                emit_event = true;
+                Some(latest.bytes_in_flight)
+            } else {
+                None
+            };
+
+        let new_ssthresh = if self.ssthresh != latest.ssthresh {
+            self.ssthresh = latest.ssthresh;
+            emit_event = true;
+            Some(latest.ssthresh)
+        } else {
+            None
+        };
+
+        if emit_event {
+            // QVis can't use all these fields and they can be large.
+            return Some(EventData::MetricsUpdated(
+                qlog::events::quic::MetricsUpdated {
+                    min_rtt: new_min_rtt,
+                    smoothed_rtt: new_smoothed_rtt,
+                    latest_rtt: new_latest_rtt,
+                    rtt_variance: new_rttvar,
+                    pto_count: None,
+                    congestion_window: new_cwnd,
+                    bytes_in_flight: new_bytes_in_flight,
+                    ssthresh: new_ssthresh,
+                    packets_in_flight: None,
+                    pacing_rate: None,
+                },
+            ));
+        }
+
+        None
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -1306,10 +1501,17 @@ mod tests {
             Ok(())
         );
 
-        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 4);
         assert_eq!(r.bytes_in_flight, 0);
 
         assert_eq!(r.lost_count, 2);
+
+        // Wait 1 RTT.
+        now += r.rtt();
+
+        r.detect_lost_packets(packet::EPOCH_APPLICATION, now, "");
+
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
     }
 
     #[test]
@@ -1459,10 +1661,17 @@ mod tests {
         r.on_loss_detection_timeout(HandshakeStatus::default(), now, "");
         assert_eq!(r.loss_probes[packet::EPOCH_APPLICATION], 0);
 
-        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 2);
         assert_eq!(r.bytes_in_flight, 0);
 
         assert_eq!(r.lost_count, 1);
+
+        // Wait 1 RTT.
+        now += r.rtt();
+
+        r.detect_lost_packets(packet::EPOCH_APPLICATION, now, "");
+
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
     }
 
     #[test]
@@ -1605,6 +1814,8 @@ mod tests {
         let mut acked = ranges::RangeSet::default();
         acked.insert(0..2);
 
+        assert_eq!(r.pkt_thresh, INITIAL_PACKET_THRESHOLD);
+
         assert_eq!(
             r.on_ack_received(
                 &acked,
@@ -1617,11 +1828,22 @@ mod tests {
             Ok(())
         );
 
-        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 4);
         assert_eq!(r.bytes_in_flight, 0);
 
         // Spurious loss.
         assert_eq!(r.lost_count, 1);
+        assert_eq!(r.lost_spurious_count, 1);
+
+        // Packet threshold was increased.
+        assert_eq!(r.pkt_thresh, 4);
+
+        // Wait 1 RTT.
+        now += r.rtt();
+
+        r.detect_lost_packets(packet::EPOCH_APPLICATION, now, "");
+
+        assert_eq!(r.sent[packet::EPOCH_APPLICATION].len(), 0);
     }
 
     #[test]
@@ -1751,11 +1973,11 @@ mod tests {
 
         // We pace this outgoing packet. as all conditions for pacing
         // are passed.
-        assert_eq!(r.pacing_rate, (12000.0 / 0.05) as u64);
+        assert_eq!(r.pacing_rate, (12000.0 * PACING_MULTIPLIER / 0.05) as u64);
         assert_eq!(
             r.get_packet_send_time().unwrap(),
             now + Duration::from_micros(
-                (6500 * 1000000) / (12000.0 / 0.05) as u64
+                (6500 * 1000000) / (12000.0 * PACING_MULTIPLIER / 0.05) as u64
             )
         );
     }
@@ -1764,4 +1986,5 @@ mod tests {
 mod cubic;
 mod delivery_rate;
 mod hystart;
+mod prr;
 mod reno;
